# type: ignore
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: arg_services/nlp/v1/nlp.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
)

import betterproto
import betterproto.lib.google.protobuf as betterproto_lib_google_protobuf
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase


if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


class SimilarityMethod(betterproto.Enum):
    """Possible methods to compute the similarity between two vectors."""

    UNSPECIFIED = 0
    """If not given, the implementation defaults to cosine similarity."""

    COSINE = 1
    """
    Cosine similarity.
    [Wikipedia](https://en.wikipedia.org/wiki/Cosine_similarity).
    """

    DYNAMAX_JACCARD = 2
    """
    DynaMax Jaccard. [Paper](https://arxiv.org/abs/1904.13264), [Code](https://
    github.com/babylonhealth/fuzzymax/blob/master/similarity/fuzzy.py).
    """

    MAXPOOL_JACCARD = 3
    """
    MaxPool Jaccard. [Paper](https://arxiv.org/abs/1904.13264), [Code](https://
    github.com/babylonhealth/fuzzymax/blob/master/similarity/fuzzy.py).
    """

    DYNAMAX_DICE = 4
    """
    DynaMax Dice. [Paper](https://arxiv.org/abs/1904.13264), [Code](https://git
    hub.com/babylonhealth/fuzzymax/blob/master/similarity/fuzzy.py).
    """

    DYNAMAX_OTSUKA = 5
    """
    DynaMax Otsuka. [Paper](https://arxiv.org/abs/1904.13264), [Code](https://g
    ithub.com/babylonhealth/fuzzymax/blob/master/similarity/fuzzy.py).
    """

    WMD = 6
    """
    Word Mover's Distance [Gensim Tutorial](https://radimrehurek.com/gensim/aut
    o_examples/tutorials/run_wmd.html).
    """

    EDIT = 7
    """
    Levenshtein distance.
    [Wikipedia](https://en.wikipedia.org/wiki/Levenshtein_distance).
    """

    JACCARD = 8
    """
    Jaccard similarity.
    [Wikipedia](https://en.wikipedia.org/wiki/Jaccard_index).
    """

    ANGULAR = 9
    """
    Angular distance.
    [Wikipedia](https://en.wikipedia.org/wiki/Angular_distance).
    """


class EmbeddingLevel(betterproto.Enum):
    UNSPECIFIED = 0
    """In the default case, no vector is computed."""

    DOCUMENT = 1
    """Compute one vector for the whole string."""

    TOKENS = 2
    """Compute vectors for all tokens in the string."""

    SENTENCES = 3
    """Compute vectors for all sentences found in the string."""


class Pooling(betterproto.Enum):
    UNSPECIFIED = 0
    """IN the default case, the arithmetic mean should be used."""

    MEAN = 1
    """
    Arithmetic mean of all elements.
    [Wikipedia](https://en.wikipedia.org/wiki/Arithmetic_mean).
    """

    MAX = 2
    """
    Maximum element of vector.
    [Wikipedia](https://en.wikipedia.org/wiki/Maximum).
    """

    MIN = 3
    """
    Minimum element of vector.
    [Wikipedia](https://en.wikipedia.org/wiki/Minimum).
    """

    SUM = 4
    """Sum of all elements."""

    FIRST = 5
    """First element of vector."""

    LAST = 6
    """Last element of vector."""

    MEDIAN = 7
    """
    Median element of vector.
    [Wikipedia](https://en.wikipedia.org/wiki/Median).
    """

    GMEAN = 8
    """
    Geometirc mean of all elements.
    [Wikipedia](https://en.wikipedia.org/wiki/Geometric_mean).
    """

    HMEAN = 9
    """
    Harmonic mean of all elements.
    [Wikipedia](https://en.wikipedia.org/wiki/Harmonic_mean).
    """


class EmbeddingType(betterproto.Enum):
    UNSPECIFIED = 0
    """In the default case, no embedding is computed."""

    SPACY = 1
    """[Spacy](https://spacy.io/models)."""

    TRANSFORMERS = 2
    """[HuggingFace Transformers](https://huggingface.co/models)."""

    SENTENCE_TRANSFORMERS = 3
    """
    [UKPLab Sentence
    Transformers](https://www.sbert.net/docs/pretrained_models.html)
    """

    TENSORFLOW_HUB = 4
    """
    Tensorflow Hub. Example: [Universal Sentence
    Encoder](https://tfhub.dev/google/universal-sentence-encoder/4).
    """

    OPENAI = 5
    """
    [OpenAI](https://platform.openai.com/docs/guides/embeddings/use-cases).
    """


@dataclass(eq=False, repr=False)
class NlpConfig(betterproto.Message):
    """Common message for configuring spacy."""

    language: str = betterproto.string_field(1)
    """
    Any language supported by spacy (e.g., `en`).
    [Reference](https://spacy.io/usage/models#languages).
    """

    spacy_model: str = betterproto.string_field(2)
    """
    Name of the trained spacy pipeline (e.g., `en_core_web_lg`). If empty, a
    blank spacy model will be used (e.g., if you only need embeddings and
    provide custom `embedding_models`. [Example: English
    models](https://spacy.io/models/en).
    """

    embedding_models: List["EmbeddingModel"] = betterproto.message_field(3)
    """
    List of embeddings to use for computing word/sentence vectors. If given,
    these embeddings will **override** the embeddings of the specified
    `spacy_model`. Multiple models are concatenated to each other, increasing
    the length of the resulting vector.
    """

    similarity_method: "SimilarityMethod" = betterproto.enum_field(4)
    """
    Mathematical function to determine a similarity score given two strings.
    """


@dataclass(eq=False, repr=False)
class SimilaritiesRequest(betterproto.Message):
    config: "NlpConfig" = betterproto.message_field(1)
    """Spacy config."""

    text_tuples: List["TextTuple"] = betterproto.message_field(2)
    """List of string pairs to compare."""

    extras: "betterproto_lib_google_protobuf.Struct" = betterproto.message_field(15)
    """Implementation-specific information can be encoded here"""


@dataclass(eq=False, repr=False)
class SimilaritiesResponse(betterproto.Message):
    similarities: List[float] = betterproto.double_field(1)
    """List of similarities ordered just like the original `text_tuples`."""

    extras: "betterproto_lib_google_protobuf.Struct" = betterproto.message_field(15)
    """Implementation-specific information can be encoded here"""


@dataclass(eq=False, repr=False)
class TextTuple(betterproto.Message):
    """Store a pair of strings."""

    text1: str = betterproto.string_field(1)
    text2: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class Strings(betterproto.Message):
    """Wrapper message to encode a list of strings that can also be `null`."""

    values: List[str] = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class DocBinRequest(betterproto.Message):
    config: "NlpConfig" = betterproto.message_field(1)
    """Spacy config."""

    texts: List[str] = betterproto.string_field(2)
    """List of strings to be processed."""

    attributes: Optional["Strings"] = betterproto.message_field(3, optional=True)
    """
    Attributes that shall be included in the DocBin object. Defaults to
    `("ORTH", "TAG", "HEAD", "DEP", "ENT_IOB", "ENT_TYPE", "ENT_KB_ID",
    "LEMMA", "MORPH", "POS")`. Possible values: `("IS_ALPHA", "IS_ASCII",
    "IS_DIGIT", "IS_LOWER", "IS_PUNCT", "IS_SPACE", "IS_TITLE", "IS_UPPER",
    "LIKE_URL", "LIKE_NUM", "LIKE_EMAIL", "IS_STOP", "IS_OOV_DEPRECATED",
    "IS_BRACKET", "IS_QUOTE", "IS_LEFT_PUNCT", "IS_RIGHT_PUNCT", "IS_CURRENCY",
    "ID", "ORTH", "LOWER", "NORM", "SHAPE", "PREFIX", "SUFFIX", "LENGTH",
    "CLUSTER", "LEMMA", "POS", "TAG", "DEP", "ENT_IOB", "ENT_TYPE", "ENT_ID",
    "ENT_KB_ID", "HEAD", "SENT_START", "SENT_END", "SPACY", "PROB", "LANG",
    "MORPH", "IDX")`. [Documentation](https://spacy.io/api/token#attributes).
    """

    enabled_pipes: "Strings" = betterproto.message_field(4, group="pipes")
    disabled_pipes: "Strings" = betterproto.message_field(5, group="pipes")
    embedding_levels: List["EmbeddingLevel"] = betterproto.enum_field(6)
    """
    List of vectors that shall be saved in the `DocBin` object. The computation
    is time-consuming, so you should only specify the embeddings you actually
    use!
    """

    extras: "betterproto_lib_google_protobuf.Struct" = betterproto.message_field(15)
    """Implementation-specific information can be encoded here"""


@dataclass(eq=False, repr=False)
class DocBinResponse(betterproto.Message):
    docbin: bytes = betterproto.bytes_field(1)
    """Serialized [`DocBin`](https://spacy.io/api/docbin) object"""

    extras: "betterproto_lib_google_protobuf.Struct" = betterproto.message_field(15)
    """Implementation-specific information can be encoded here"""


@dataclass(eq=False, repr=False)
class VectorsRequest(betterproto.Message):
    config: "NlpConfig" = betterproto.message_field(1)
    """Spacy config."""

    texts: List[str] = betterproto.string_field(2)
    """List of strings that shall be embedded (i.e., converted to vectors)."""

    embedding_levels: List["EmbeddingLevel"] = betterproto.enum_field(3)
    """
    List of vectors that shall be returned. The computation is time-consuming,
    so you should only specify the embeddings you actually use!
    """

    extras: "betterproto_lib_google_protobuf.Struct" = betterproto.message_field(15)
    """Implementation-specific information can be encoded here"""


@dataclass(eq=False, repr=False)
class VectorsResponse(betterproto.Message):
    vectors: List["VectorResponse"] = betterproto.message_field(1)
    """List of vectors whose order corresponds to the one of `texts`."""

    extras: "betterproto_lib_google_protobuf.Struct" = betterproto.message_field(15)
    """Implementation-specific information can be encoded here"""


@dataclass(eq=False, repr=False)
class VectorResponse(betterproto.Message):
    """
    Container object that includes vectors for all levels specified in
    `embedding_levels`.
    """

    document: "Vector" = betterproto.message_field(1)
    """One vector for the whole string."""

    tokens: List["Vector"] = betterproto.message_field(2)
    """Vectors for all tokens in the string."""

    sentences: List["Vector"] = betterproto.message_field(3)
    """Vectors for all sentences found in the string."""

    extras: "betterproto_lib_google_protobuf.Struct" = betterproto.message_field(15)
    """Implementation-specific information can be encoded here"""


@dataclass(eq=False, repr=False)
class Vector(betterproto.Message):
    """Container for storing a vector as a list of floats."""

    vector: List[float] = betterproto.double_field(1)


@dataclass(eq=False, repr=False)
class EmbeddingModel(betterproto.Message):
    """
    Specification of one model that is used to generate embeddings for strings.
    """

    model_type: "EmbeddingType" = betterproto.enum_field(1)
    """
    Each embedding has to be implemented, thus this enum is used to select the
    correct one.
    """

    model_name: str = betterproto.string_field(2)
    """
    You have to specify the name of the model that should be used by the
    selected impelemtation (i.e., `model_type`). We provide links to exemplary
    models for each implementation in the documentation of `EmbeddingType`.
    """

    pooling_type: "Pooling" = betterproto.enum_field(3, group="pooling")
    """Standard pooling functions like mean, min, max."""

    pmean: float = betterproto.double_field(4, group="pooling")
    """
    Power mean (or generalized mean). This method allows you to alter the
    computation of the mean representation. Special cases include arithmetic
    mean (p = 1), geometric mean (p = 0), harmonic mean (p = -1), minimum (p =
    -∞), maximum (p = ∞).
    [Wikipedia](https://en.wikipedia.org/wiki/Generalized_mean).
    [Paper](https://arxiv.org/abs/1803.01400).
    """


class NlpServiceStub(betterproto.ServiceStub):
    async def vectors(
        self,
        vectors_request: "VectorsRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "VectorsResponse":
        return await self._unary_unary(
            "/arg_services.nlp.v1.NlpService/Vectors",
            vectors_request,
            VectorsResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def similarities(
        self,
        similarities_request: "SimilaritiesRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "SimilaritiesResponse":
        return await self._unary_unary(
            "/arg_services.nlp.v1.NlpService/Similarities",
            similarities_request,
            SimilaritiesResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def doc_bin(
        self,
        doc_bin_request: "DocBinRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "DocBinResponse":
        return await self._unary_unary(
            "/arg_services.nlp.v1.NlpService/DocBin",
            doc_bin_request,
            DocBinResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class NlpServiceBase(ServiceBase):
    async def vectors(self, vectors_request: "VectorsRequest") -> "VectorsResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def similarities(
        self, similarities_request: "SimilaritiesRequest"
    ) -> "SimilaritiesResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def doc_bin(self, doc_bin_request: "DocBinRequest") -> "DocBinResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_vectors(
        self, stream: "grpclib.server.Stream[VectorsRequest, VectorsResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.vectors(request)
        await stream.send_message(response)

    async def __rpc_similarities(
        self, stream: "grpclib.server.Stream[SimilaritiesRequest, SimilaritiesResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.similarities(request)
        await stream.send_message(response)

    async def __rpc_doc_bin(
        self, stream: "grpclib.server.Stream[DocBinRequest, DocBinResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.doc_bin(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/arg_services.nlp.v1.NlpService/Vectors": grpclib.const.Handler(
                self.__rpc_vectors,
                grpclib.const.Cardinality.UNARY_UNARY,
                VectorsRequest,
                VectorsResponse,
            ),
            "/arg_services.nlp.v1.NlpService/Similarities": grpclib.const.Handler(
                self.__rpc_similarities,
                grpclib.const.Cardinality.UNARY_UNARY,
                SimilaritiesRequest,
                SimilaritiesResponse,
            ),
            "/arg_services.nlp.v1.NlpService/DocBin": grpclib.const.Handler(
                self.__rpc_doc_bin,
                grpclib.const.Cardinality.UNARY_UNARY,
                DocBinRequest,
                DocBinResponse,
            ),
        }



import itertools
import logging
import tarfile
import typing as t
import urllib.request
import warnings
from abc import ABC, abstractmethod
from collections import abc
from dataclasses import dataclass
from pathlib import Path

import numpy as np
import scipy.stats
import spacy
from mashumaro.mixins.dict import DataClassDictMixin
from rich.progress import Progress
from spacy import about as spacy_about
from spacy.cli.download import (
    get_latest_version as spacy_get_latest_version,
)
from spacy.cli.download import (
    get_model_filename as spacy_get_model_filename,
)
from spacy.language import Language as SpacyLanguage
from spacy.tokens import Doc
from thinc.types import Floats1d as SpacyVector

from nlp_service import nlp_pb
from nlp_service.sim_funcs import SimilarityFactory as SimilarityFactory
from nlp_service.typing import ArrayLike, NumpyMatrix, NumpyVector

log = logging.getLogger(__name__)

__all__ = ("docs", "doc", "vectors", "vector", "similarities", "similarity")

CACHE_FOLDER = Path.home() / ".cache" / "nlp-service"

torch_device = "cpu"

try:
    from torch.cuda import is_available as is_cuda_available

    torch_device = "cuda" if is_cuda_available() else "cpu"
    log.info(f"Using torch device '{torch_device}'.")
except ModuleNotFoundError:
    log.info("'torch' not installed.")

# https://spacy.io/usage/processing-pipelines#built-in
spacy_components: tuple[str, ...] = (
    "tagger",
    "parser",
    "ner",
    "entity_linker",
    "entity_ruler",
    "textcat",
    "textcat_multilabel",
    "lemmatizer",
    "morphologizer",
    "attribute_ruler",
    "senter",
    "sentencizer",
    # tok2vec, transformer
)
custom_components: tuple[str, ...] = ("embeddings_factory", "similarity_factory")


@dataclass(frozen=True, eq=True)
class SerializableEmbeddingModel(DataClassDictMixin):
    model_type: int
    model_name: str
    pooling_type: int
    pmean: float

    @classmethod
    def from_protobuf(cls, pb: nlp_pb.EmbeddingModel) -> "SerializableEmbeddingModel":
        return cls(pb.model_type, pb.model_name, pb.pooling_type, pb.pmean)


class ModelBase(ABC):
    @abstractmethod
    def __init__(self, model: SerializableEmbeddingModel) -> None:
        pass

    @abstractmethod
    def vector(self, text: str) -> SpacyVector:
        pass


def pmean(vectors: ArrayLike, p: float) -> SpacyVector:
    return np.power(
        np.mean(np.power(np.array(vectors, dtype=complex), p), axis=0), 1 / p
    ).real


class SpacyModel(ModelBase):
    def __init__(self, model: SerializableEmbeddingModel):
        self.model = spacy.load(model.model_name)
        self.pooling_type = model.pooling_type
        self.pmean = model.pmean

    def vector(self, text: str) -> SpacyVector:
        with self.model.select_pipes(enable=["senter"]):
            doc = self.model(text)

        if len(doc) > 1:
            if self.pooling_type and self.pooling_type != nlp_pb.Pooling.MEAN:
                return t.cast(
                    SpacyVector,
                    pool_map[self.pooling_type](
                        np.array([token.vector for token in doc])
                    ),
                )
            elif self.pmean:
                return pmean(
                    [t.cast(NumpyVector, token.vector) for token in doc], self.pmean
                )

        return doc.vector


embedding_map: dict[
    nlp_pb.EmbeddingType, t.Callable[[SerializableEmbeddingModel], ModelBase]
] = {
    nlp_pb.EmbeddingType.SPACY: SpacyModel,
}


try:
    import torch
    from transformers import AutoModel, AutoTokenizer

    class TransformersModel(ModelBase):
        def __init__(self, model: SerializableEmbeddingModel):
            # Load model from HuggingFace Hub
            try:
                self.tokenizer = AutoTokenizer.from_pretrained(
                    model.model_name, use_fast=True
                )
            except Exception:
                self.tokenizer = AutoTokenizer.from_pretrained(model.model_name)

            self.model = AutoModel.from_pretrained(model.model_name).to(torch_device)

        # Mean nlp_pb.Pooling - Take attention mask into account for correct averaging
        def mean_pooling(self, model_output, attention_mask) -> torch.Tensor:
            # First element of model_output contains all token embeddings
            token_embeddings = model_output[0]
            input_mask_expanded = (
                attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
            )
            return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(
                input_mask_expanded.sum(1), min=1e-9
            )

        def vector(self, text: str) -> SpacyVector:
            # Tokenize sentences
            encoded_input = self.tokenizer(
                text, padding=True, truncation=True, return_tensors="pt"
            )
            encoded_input = encoded_input.to(torch_device)
            # Compute token embeddings
            with torch.no_grad():
                model_output = self.model(**encoded_input)
            # Perform pooling
            sentence_embeddings = self.mean_pooling(
                model_output, encoded_input["attention_mask"]
            )
            # Normalize embeddings; is this needed? Normilize? Logit?
            # sentence_embeddings = F.normalize(sentence_embeddings, p=2, dim=1)
            return sentence_embeddings[0].cpu().numpy()

    embedding_map[nlp_pb.EmbeddingType.TRANSFORMERS] = TransformersModel

except ModuleNotFoundError:
    log.info("'transformers' not installed.")


try:
    from sentence_transformers import SentenceTransformer

    class SentenceTransformersModel(ModelBase):
        def __init__(self, model: SerializableEmbeddingModel):
            self.model = SentenceTransformer(model.model_name, device=torch_device)

        def vector(self, text: str) -> SpacyVector:
            embeddings = t.cast(
                NumpyVector, self.model.encode([text], convert_to_numpy=True)
            )

            return embeddings[0]

    embedding_map[nlp_pb.EmbeddingType.SENTENCE_TRANSFORMERS] = (
        SentenceTransformersModel
    )

except ModuleNotFoundError:
    log.info("'sentence-transformers' not installed.")


try:
    import openai

    class OpenaiModel(ModelBase):
        def __init__(self, model: SerializableEmbeddingModel):
            self.model_name: str = model.model_name
            self.client = openai.Client()

        def vector(self, text: str) -> SpacyVector:
            res = self.client.embeddings.create(input=[text], model=self.model_name)

            return t.cast(SpacyVector, np.array(res.data[0].embedding))

    embedding_map[nlp_pb.EmbeddingType.OPENAI] = OpenaiModel

except ModuleNotFoundError:
    log.info("'openai' not installed.")


@SpacyLanguage.factory("embeddings_factory")
class EmbeddingsFactory:
    def __init__(self, nlp, name, models):
        self.models: list[ModelBase] = []

        for model_dict in models:
            model = SerializableEmbeddingModel.from_dict(model_dict)
            model_class = embedding_map.get(model.model_type)

            if model_class is None:
                raise ValueError(
                    "The packages required for"
                    f" '{model.model_type.name}' are not"
                    " installed"
                )

            if model not in model_cache:
                model_cache[model] = model_class(model)

            self.models.append(model_cache[model])

    def __call__(self, doc):
        if len(self.models) > 0:
            doc.user_hooks["vector"] = self.vector
            doc.user_span_hooks["vector"] = self.vector
            doc.user_token_hooks["vector"] = self.vector

        return doc

    def vector(self, obj) -> SpacyVector:
        vecs = np.array([model.vector(obj.text) for model in self.models])
        return t.cast(SpacyVector, np.concatenate(vecs))


SpacyKey = tuple[str, str, tuple[SerializableEmbeddingModel, ...]]
SpacyCache = t.Tuple[SpacyLanguage, dict[str, Doc]]
spacy_cache: dict[SpacyKey, SpacyCache] = {}
model_cache: dict[SerializableEmbeddingModel, ModelBase] = {}


class UrlReportHook:
    def __init__(self, progress: Progress, name: str):
        self.task = None
        self.progress = progress
        self.name = name

    def __call__(self, block_num: int, block_size: int, total_size: int):
        if self.task is None:
            self.task = self.progress.add_task(
                f"Downloading {self.name}...", total=total_size
            )

        downloaded = block_num * block_size

        if downloaded < total_size:
            self.progress.update(self.task, completed=downloaded)

        if self.progress.finished:
            self.task = None


def get_tarfile_members(
    tf: tarfile.TarFile, prefix: str
) -> t.Generator[tarfile.TarInfo, None, None]:
    prefix_len = len(prefix)

    for member in tf.getmembers():
        if member.path.startswith(prefix):
            member.path = member.path[prefix_len:]
            yield member


def _load_spacy_model(name: str | None) -> SpacyLanguage:
    if not name:
        return spacy.blank("en")

    version = spacy_get_latest_version(name)
    filename = spacy_get_model_filename(name, version, sdist=True)
    versioned_name = f"{name}-{version}"
    path = CACHE_FOLDER / "spacy" / versioned_name
    tmpfile = path.with_suffix(".tar.gz")

    if not path.exists():
        path.parent.mkdir(parents=True, exist_ok=True)
        download_url = f"{spacy_about.__download_url__}/{filename}"
        with Progress() as progress:
            urllib.request.urlretrieve(
                download_url, tmpfile, UrlReportHook(progress, versioned_name)
            )

        with tarfile.open(tmpfile, mode="r:gz") as tf:
            member_prefix = f"{versioned_name}/{name}/{versioned_name}/"
            members = get_tarfile_members(tf, member_prefix)
            tf.extractall(path=path, members=members)

        tmpfile.unlink()

    return spacy.load(path)


def load_spacy(config: nlp_pb.NlpConfig) -> SpacyCache:
    models = tuple(
        SerializableEmbeddingModel.from_protobuf(model)
        for model in config.embedding_models
    )
    key: SpacyKey = (
        config.language,
        config.spacy_model,
        models,
    )

    if key not in spacy_cache:
        nlp = _load_spacy_model(config.spacy_model)

        if models:
            nlp.add_pipe(
                "embeddings_factory",
                last=True,
                config={"models": [model.to_dict() for model in models]},
            )

        if config.similarity_method not in [
            nlp_pb.SimilarityMethod.COSINE,
            nlp_pb.SimilarityMethod.UNSPECIFIED,
        ]:
            nlp.add_pipe(
                "similarity_factory",
                last=True,
                config={"method": config.similarity_method},
            )

        spacy_cache[key] = (nlp, {})

    return spacy_cache[key]


pool_map: dict[nlp_pb.Pooling, t.Callable[[NumpyMatrix], NumpyVector]] = {
    nlp_pb.Pooling.MEAN: lambda x: np.mean(x, axis=0),
    nlp_pb.Pooling.FIRST: lambda x: x[0],
    nlp_pb.Pooling.LAST: lambda x: x[-1],
    nlp_pb.Pooling.MIN: lambda x: np.min(x, axis=0),
    nlp_pb.Pooling.MAX: lambda x: np.max(x, axis=0),
    nlp_pb.Pooling.SUM: lambda x: np.sum(x, axis=0),
    nlp_pb.Pooling.MEDIAN: lambda x: np.median(x, axis=0),
    nlp_pb.Pooling.GMEAN: lambda x: scipy.stats.gmean(x, axis=0),
    nlp_pb.Pooling.HMEAN: lambda x: scipy.stats.hmean(x, axis=0),
}


class PipeSelection(t.TypedDict, total=False):
    enable: list[str] | tuple[str, ...]
    disable: list[str] | tuple[str, ...]


def docs(
    texts: abc.Iterable[str],
    config: nlp_pb.NlpConfig,
    pipes_selection: PipeSelection | None = None,
) -> list[Doc]:
    if not pipes_selection:
        pipes_selection = {"disable": []}  # if empty, spacy will raise an exception

    if "enable" in pipes_selection:
        pipes_selection["enable"] = tuple(pipes_selection["enable"]) + custom_components

    # arg_services.require_all(["config.language"], req, ctx)

    # for model in req.config.embedding_models:
    #     arg_services.require_all(
    #         ["model_type", "model_name"],
    #         model,
    #         ctx,
    #         parent="embeddings_factory",
    #     )

    # TODO: Cache not used due to the ability to enable/disable certain pipes
    nlp, _ = load_spacy(config)

    with nlp.select_pipes(**pipes_selection):
        return list(nlp.pipe(texts))


def doc(
    text: str,
    config: nlp_pb.NlpConfig,
    pipes_selection: PipeSelection | None = None,
) -> Doc:
    return docs([text], config, pipes_selection)[0]


EmbeddingLevelSingle = t.Literal[
    nlp_pb.EmbeddingLevel.DOCUMENT, nlp_pb.EmbeddingLevel.UNSPECIFIED
]
EmbeddingLevelMulti = t.Literal[
    nlp_pb.EmbeddingLevel.SENTENCES, nlp_pb.EmbeddingLevel.TOKENS
]


@t.overload
def vectors(
    texts: abc.Iterable[str],
    config: nlp_pb.NlpConfig,
    embedding_level: EmbeddingLevelSingle,
) -> list[SpacyVector]: ...


@t.overload
def vectors(
    texts: abc.Iterable[str],
    config: nlp_pb.NlpConfig,
    embedding_level: EmbeddingLevelMulti,
) -> list[list[SpacyVector]]: ...


@t.overload
def vectors(
    texts: abc.Iterable[str],
    config: nlp_pb.NlpConfig,
    embedding_level: nlp_pb.EmbeddingLevel,
) -> list[SpacyVector] | list[list[SpacyVector]]: ...


def vectors(
    texts: abc.Iterable[str],
    config: nlp_pb.NlpConfig,
    embedding_level: nlp_pb.EmbeddingLevel = nlp_pb.EmbeddingLevel.UNSPECIFIED,
) -> list[SpacyVector] | list[list[SpacyVector]]:
    # arg_services.require_all(["config.language", "embedding_levels"], req, ctx)
    # arg_services.require_all_repeated(
    #     "config.embedding_models", ["model_type", "model_name"], req, ctx
    # )

    nlp, doc_cache = load_spacy(config)

    if new_texts := [text for text in texts if text not in doc_cache]:
        with nlp.select_pipes(enable=custom_components):
            doc_cache.update(zip(new_texts, nlp.pipe(new_texts)))

    if embedding_level in {
        nlp_pb.EmbeddingLevel.DOCUMENT,
        nlp_pb.EmbeddingLevel.UNSPECIFIED,
    }:
        doc_vecs: list[SpacyVector] = []

        for text in texts:
            doc = doc_cache[text]
            doc_vecs.append(doc.vector)

        return doc_vecs
    else:
        span_vecs: list[list[SpacyVector]] = []

        for text in texts:
            doc = doc_cache[text]

            if embedding_level == nlp_pb.EmbeddingLevel.TOKENS:
                span_vecs.append([token.vector for token in doc])
            elif embedding_level == nlp_pb.EmbeddingLevel.SENTENCES:
                span_vecs.append([sent.vector for sent in doc.sents])

        return span_vecs


@t.overload
def vector(
    text: str, config: nlp_pb.NlpConfig, embedding_level: EmbeddingLevelSingle
) -> SpacyVector: ...


@t.overload
def vector(
    text: str, config: nlp_pb.NlpConfig, embedding_level: EmbeddingLevelMulti
) -> list[SpacyVector]: ...


@t.overload
def vector(
    text: str, config: nlp_pb.NlpConfig, embedding_level: nlp_pb.EmbeddingLevel
) -> SpacyVector | list[SpacyVector]: ...


def vector(
    text: str,
    config: nlp_pb.NlpConfig,
    embedding_level: nlp_pb.EmbeddingLevel = nlp_pb.EmbeddingLevel.UNSPECIFIED,
) -> SpacyVector | list[SpacyVector]:
    return vectors([text], config, embedding_level)[0]


def similarities(
    text_tuples: abc.Iterable[tuple[str, str]], config: nlp_pb.NlpConfig
) -> list[float]:
    # arg_services.require_all(["config.language", "config.similarity_method"], req, ctx)
    # arg_services.require_all_repeated(
    #     "config.embedding_models", ["model_type", "model_name"], req, ctx
    # )
    # arg_services.require_all_repeated("text_tuples", ["text1", "text2"], req, ctx)

    nlp, doc_cache = load_spacy(config)
    texts = itertools.chain.from_iterable((x[0], x[1]) for x in text_tuples)

    if new_texts := [text for text in texts if text not in doc_cache]:
        with nlp.select_pipes(enable=custom_components):
            doc_cache.update(zip(new_texts, nlp.pipe(new_texts)))

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        return [doc_cache[x[0]].similarity(doc_cache[x[1]]) for x in text_tuples]


def similarity(text1: str, text2: str, config: nlp_pb.NlpConfig) -> float:
    return similarities([(text1, text2)], config)[0]



import asyncio
import logging
import shutil
import sys
from enum import Enum
from pathlib import Path
from typing import Optional

import typer
from rich import print

from nlp_service import lib, nlp_pb, server
from nlp_service.sim_funcs import SimilarityFactory as SimilarityFactory

log = logging.getLogger(__name__)

app = typer.Typer()


class NlpConfig(str, Enum):
    default = "default"
    strf = "strf"
    openai = "openai"


NLP_CONFIG: dict[NlpConfig, nlp_pb.NlpConfig] = {
    NlpConfig.default: nlp_pb.NlpConfig(
        language="en",
        spacy_model="en_core_web_lg",
        similarity_method=nlp_pb.SimilarityMethod.COSINE,
    ),
    NlpConfig.strf: nlp_pb.NlpConfig(
        language="en",
        similarity_method=nlp_pb.SimilarityMethod.COSINE,
        embedding_models=[
            nlp_pb.EmbeddingModel(
                model_type=nlp_pb.EmbeddingType.TRANSFORMERS,
                model_name="multi-qa-MiniLM-L6-cos-v1",
                pooling_type=nlp_pb.Pooling.MEAN,
            )
        ],
    ),
    NlpConfig.openai: nlp_pb.NlpConfig(
        language="en",
        similarity_method=nlp_pb.SimilarityMethod.COSINE,
        embedding_models=[
            nlp_pb.EmbeddingModel(
                model_type=nlp_pb.EmbeddingType.OPENAI,
                model_name="text-embedding-ada-002",
                pooling_type=nlp_pb.Pooling.MEAN,
            )
        ],
    ),
}


@app.command()
def clear_cache():
    shutil.rmtree(lib.CACHE_FOLDER)


@app.command()
def retrieve(
    cases_folder: Path,
    query: Optional[str] = None,
    query_file: Optional[Path] = None,
    include_pattern: str = "**/*.txt",
    exclude_pattern: str = "",
    show_contents: bool = False,
    limit: int = sys.maxsize,
    config: NlpConfig = NlpConfig.default,
):
    """Find the most similar cases to a given query."""

    assert (
        query is not None or query_file is not None
    ), "Either query or query_file must be provided"
    assert not (
        query is not None and query_file is not None
    ), "Only one of query or query_file must be provided"

    query_name = "stdin"
    query_text = query or ""

    if query_file is not None:
        query_name = query_file.name
        query_text = query_file.read_text()

    print("Retrieving ranking...")
    print(f"Query '{query_name}'")

    if show_contents:
        print(f"{query_text}")
        print()

    cases = {
        str(case_file.relative_to(cases_folder)): case_file.read_text()
        for case_file in cases_folder.glob(include_pattern)
        if not case_file.match(exclude_pattern)
    }

    sim_values = lib.similarities(
        [(query_name, case_name) for case_name in cases.keys()], NLP_CONFIG[config]
    )
    sim_tuples = zip(cases, sim_values)
    ranking = sorted(sim_tuples, key=lambda x: x[1], reverse=True)

    for i, (case_name, sim) in enumerate(ranking[:limit], start=1):
        print(f"{i}. '{case_name}' ({sim:.3f})")

        if show_contents:
            print(f"{cases[case_name]}")
            print()


@app.command()
def serve(host: str = "127.0.0.1", port: int = 50100):
    """Main entry point for the server."""
    asyncio.run(server.main(host, port))


if __name__ == "__main__":
    app()



import numpy as np
import spacy
from spacy.language import Language as SpacyLanguage
from spacy.tokens import Doc, DocBin, Span, Token

from nlp_service import nlp_pb
from nlp_service.sim_funcs import SimilarityFactory as SimilarityFactory
from nlp_service.typing import ArrayLike, NumpyVector

Doc.set_extension("vector", default=None)
Span.set_extension("vector", default=None)
Token.set_extension("vector", default=None)


def blank(
    language: str,
    similarity_method: nlp_pb.SimilarityMethod = nlp_pb.SimilarityMethod.UNSPECIFIED,
) -> SpacyLanguage:
    spacy_lang = spacy.blank(language)
    inject_pipes(spacy_lang, similarity_method)

    return spacy_lang


def docbin2docs(
    docbin_bytes: bytes,
    language: str | SpacyLanguage,
    similarity_method: nlp_pb.SimilarityMethod = nlp_pb.SimilarityMethod.UNSPECIFIED,
) -> tuple[Doc, ...]:
    if isinstance(language, str):
        language = blank(language, similarity_method)

    docbin = DocBin().from_bytes(docbin_bytes)

    return tuple(docbin.get_docs(language.vocab))


def list2array(values: ArrayLike) -> NumpyVector:
    return np.array(values)


def inject_vectors(
    doc: Doc,
    res: nlp_pb.VectorResponse,
) -> None:
    if res.document:
        doc._.set("vector", list2array(res.document.vector))

    if res.sentences:
        for sent, sent_res in zip(doc.sents, res.sentences):
            sent._.set("vector", list2array(sent_res.vector))

    if res.tokens:
        for token, token_res in zip(doc, res.tokens):
            token._.set("vector", list2array(token_res.vector))


def inject_pipes(
    nlp: SpacyLanguage,
    similarity_method: nlp_pb.SimilarityMethod = nlp_pb.SimilarityMethod.UNSPECIFIED,
) -> None:
    nlp.add_pipe("remote_vector", last=True)
    nlp.add_pipe("similarity_factory", last=True, config={"method": similarity_method})


@SpacyLanguage.component("remote_vector")
def remote_vector(doc):
    def func(x):
        return x._.vector

    doc.user_hooks["vector"] = func
    doc.user_span_hooks["vector"] = func
    doc.user_token_hooks["vector"] = func

    return doc



import logging

import betterproto
from grpclib.server import Server
from grpclib.utils import graceful_exit
from spacy.tokens import DocBin

from nlp_service import lib, nlp_pb
from nlp_service.sim_funcs import SimilarityFactory as SimilarityFactory

log = logging.getLogger(__name__)


class NlpService(nlp_pb.NlpServiceBase):
    async def doc_bin(self, req: nlp_pb.DocBinRequest) -> nlp_pb.DocBinResponse:
        res = nlp_pb.DocBinResponse()

        pipes_selection: lib.PipeSelection | None = None

        if betterproto.which_one_of(req, "pipes") == "enabled_pipes":
            pipes_selection = {"enable": tuple(req.enabled_pipes.values)}
        elif betterproto.which_one_of(req, "pipes") == "disabled_pipes":
            pipes_selection = {"disable": tuple(req.disabled_pipes.values)}

        docs = lib.docs(req.texts, req.config, pipes_selection)

        if levels := req.embedding_levels:
            for doc in docs:
                if nlp_pb.EmbeddingLevel.DOCUMENT in levels:
                    doc._.set("vector", doc.vector)
                if nlp_pb.EmbeddingLevel.TOKENS in levels:
                    for token in doc:
                        token._.set("vector", token.vector)
                if nlp_pb.EmbeddingLevel.SENTENCES in levels:
                    for sent in doc.sents:
                        sent._.set("vector", sent.vector)

        if req.attributes is not None:
            res.docbin = DocBin(
                req.attributes.values, docs=docs, store_user_data=True
            ).to_bytes()
        else:
            res.docbin = DocBin(docs=docs, store_user_data=True).to_bytes()

        return res

    async def vectors(self, req: nlp_pb.VectorsRequest) -> nlp_pb.VectorsResponse:
        res = nlp_pb.VectorsResponse(
            vectors=[nlp_pb.VectorResponse() for _ in req.texts]
        )

        for level in req.embedding_levels:
            if (
                level == nlp_pb.EmbeddingLevel.DOCUMENT
                or level == nlp_pb.EmbeddingLevel.UNSPECIFIED
            ):
                text_vectors = lib.vectors(req.texts, req.config, level)

                for i, vector in enumerate(text_vectors):
                    res.vectors[i].document.vector.extend(vector.tolist())

            elif level == nlp_pb.EmbeddingLevel.TOKENS:
                text_vectors = lib.vectors(req.texts, req.config, level)

                for i, vectors in enumerate(text_vectors):
                    res.vectors[i].tokens.extend(
                        nlp_pb.Vector(vector=vector.tolist()) for vector in vectors
                    )

            elif level == nlp_pb.EmbeddingLevel.SENTENCES:
                text_vectors = lib.vectors(req.texts, req.config, level)

                for i, vectors in enumerate(text_vectors):
                    res.vectors[i].tokens.extend(
                        nlp_pb.Vector(vector=vector.tolist()) for vector in vectors
                    )

        return res

    async def similarities(
        self, req: nlp_pb.SimilaritiesRequest
    ) -> nlp_pb.SimilaritiesResponse:
        text_tuples = [(x.text1, x.text2) for x in req.text_tuples]

        return nlp_pb.SimilaritiesResponse(
            similarities=lib.similarities(text_tuples, req.config)
        )


async def main(host: str, port: int):
    server = Server([NlpService()])
    with graceful_exit([server]):
        await server.start(host, port)
        print(f"Serving on {host}:{port}")
        await server.wait_closed()



import nptyping as npt
from numpy.typing import ArrayLike as ArrayLike
from spacy.tokens import Doc, Span, Token
from thinc.types import Floats1d

SpacyObj = Doc | Token | Span
NumpyVector = npt.NDArray[npt.Shape["*"], npt.Floating]
NumpyMatrix = npt.NDArray[npt.Shape["*, ..."], npt.Floating]
SpacyVector = Floats1d



import logging

from . import client, sim_funcs, typing
from .lib import doc, docs, similarities, similarity, vector, vectors
from .nlp_pb import EmbeddingModel, EmbeddingType, NlpConfig, Pooling, SimilarityMethod

__all__ = (
    "client",
    "sim_funcs",
    "typing",
    "docs",
    "doc",
    "vectors",
    "vector",
    "similarities",
    "similarity",
    "NlpConfig",
    "EmbeddingModel",
    "SimilarityMethod",
    "EmbeddingType",
    "Pooling",
)

logging.getLogger(__name__).addHandler(logging.NullHandler())
